from datetime import date
from unittest import TestCase
from unittest import mock

from personnummer import personnummer
import urllib.request
import json


def get_test_data():
    response = urllib.request.urlopen(
        'https://raw.githubusercontent.com/personnummer/meta/master/testdata/list.json')
    raw = response.read().decode('utf-8')
    return json.loads(raw)


test_data = get_test_data()
availableListFormats = [
    'long_format',
    'short_format',
    'separated_format',
    'separated_long',
]


class TestPersonnummer(TestCase):
    def testPersonnummerList(self):
        for item in test_data:
            for available_format in availableListFormats:
                self.assertEqual(personnummer.valid(
                    item[available_format]), item['valid'])

    def testPersonnummerFormat(self):
        for item in test_data:
            if not item['valid']:
                continue

            expected_long_format = item['long_format']
            expected_separated_format: str = item['separated_format']
            for available_format in availableListFormats:
                if available_format == 'short_format' and '+' in expected_separated_format:
                    # Since the short format is missing the separator,
                    # the library will never use the `+` separator
                    # in the outputted format
                    continue
                self.assertEqual(
                    expected_separated_format,
                    personnummer.parse(item[available_format]).format()
                )
                self.assertEqual(
                    expected_long_format,
                    personnummer.parse(item[available_format]).format(True)
                )

    def testPersonnummerError(self):
        for item in test_data:
            if item['valid']:
                continue

            for available_format in availableListFormats:
                self.assertRaises(
                    personnummer.PersonnummerException,
                    personnummer.parse,
                    item[available_format],
                )

    def testPersonnummerSex(self):
        for item in test_data:
            if not item['valid']:
                continue

            for available_format in availableListFormats:
                self.assertEqual(personnummer.parse(
                    item[available_format]).is_male(), item['isMale'])
                self.assertEqual(personnummer.parse(
                    item[available_format]).is_female(), item['isFemale'])

    def testPersonnummerAge(self):
        for item in test_data:
            if not item['valid']:
                continue

            separated_format = item['separated_format']
            pin = item['separated_long']
            year = int(pin[0:4])
            month = int(pin[4:6])
            day = int(pin[6:8])

            if item['type'] == 'con':
                day -= 60
            if '+' in separated_format:
                # This is needed in order for the age to be the same
                # when testing the 'long_format' and any of the separated_*
                # formats. Otherwise, the long format will have an age of 0
                # and the separated ones will have an age of 100.
                year += 100

            mocked_date = date(year=year, month=month, day=day)
            for available_format in availableListFormats:
                if available_format == 'short_format' and '+' in separated_format:
                    # Since the short format is missing the separator,
                    # the library will never use the `+` separator
                    # in the outputted format
                    continue
                p = personnummer.parse(item[available_format])
                with mock.patch(
                    'personnummer.personnummer._get_current_date',
                    mock.Mock(return_value=mocked_date)
                ):
                    if '+' in separated_format:
                        self.assertEqual(100, p.get_age())
                    else:
                        self.assertEqual(0, p.get_age())



import argparse

from personnummer import personnummer
from personnummer.personnummer import PersonnummerException

def setup_args():
    ap = argparse.ArgumentParser('personnummer')
    ap.add_argument('pnr')
    return ap


def main():
    ap = setup_args()
    args = ap.parse_args()
    if args.pnr:
        try:
            pn = personnummer.Personnummer(args.pnr)
            
            print(f'Age: {pn.get_age()}')
            sex = 'female'
            if pn.is_male():
                sex = 'male'
            print(f'Sex: {sex}')

        except PersonnummerException:
            print('Not a valid Swedish personnummer')
            




import datetime
import math
import re

string_types = str


class PersonnummerException(Exception):
    pass


class PersonnummerInvalidException(PersonnummerException):
    pass


class PersonnummerParseException(PersonnummerException):
    pass


class Personnummer:
    def __init__(self, ssn, options=None):
        """
        Initializes the Object and checks if the given Swedish personal identity number is valid.
        :param ssn
        :type ssn str
        :param options
        :type options dict
        """

        if options is None:
            options = {}

        self.options = options
        self._ssn = ssn
        self._parse_parts(ssn)
        self._validate()

    @property
    def parts(self) -> dict:
        return {
            'century': self.century,
            'year': self.year,
            'month': self.month,
            'day': self.day,
            'sep': self.sep,
            'num': self.num,
            'check': self.check,
        }

    def is_coordination_number(self):
        return int(self.day) > 60

    def format(self, long_format=False):
        """
        Format a Swedish personal identity number as one of the official formats,
        A long format or a short format.

        This function raises a ValueError if the input number could not be parsed
        as a valid Swedish personal identity number

        :param long_format: Defaults to False and formats a personal identity number
            as YYMMDD-XXXX. If set to True the format will be YYYYMMDDXXXX.
        :type long_format: bool
        :rtype: str
        :return:
        """

        if long_format:
            ssn_format = '{century}{year}{month}{day}{num}{check}'
        else:
            ssn_format = '{year}{month}{day}{sep}{num}{check}'

        return ssn_format.format(
            century=self.century,
            year=self.year,
            month=self.month,
            day=self.day,
            sep=self.sep,
            num=self.num,
            check=self.check,
        )

    def get_date(self):
        """
        Get the underlying date from a social security number

        :rtype: datetime.date
        """
        year = int(self.full_year)
        month = int(self.month)
        day = int(self.day)
        day = day - 60 if self.is_coordination_number() else day
        return datetime.date(year, month, day)

    def get_age(self):
        """
        Get the age of a person from a Swedish personal identity number

        :rtype: int
        :return:
        """
        today = _get_current_date()

        year = int(self.full_year)
        month = int(self.month)
        day = int(self.day)
        day = day - 60 if self.is_coordination_number() else day

        return today.year - year - ((today.month, today.day) < (month, day))

    def is_female(self):
        return not self.is_male()

    def is_male(self):
        gender_digit = int(self.num)

        return gender_digit % 2 != 0

    def _parse_parts(self, ssn):
        """
        Get different parts of a Swedish personal identity number
        :param ssn
        :type ssn str|int
        """
        reg = r"^(\d{2}){0,1}(\d{2})(\d{2})(\d{2})([\-\+]{0,1})?((?!000)\d{3})(\d{0,1})$"
        match = re.match(reg, str(ssn))

        if not match:
            raise PersonnummerParseException(
                'Could not parse "{}" as a valid Swedish SSN.'.format(ssn))

        century = match.group(1)
        year = match.group(2)
        month = match.group(3)
        day = match.group(4)
        sep = match.group(5)
        num = match.group(6)
        check = match.group(7)

        if not century:
            base_year = _get_current_date().year
            if sep == '+':
                base_year -= 100
            else:
                sep = '-'
            full_year = base_year - ((base_year - int(year)) % 100)
            century = str(int(full_year / 100))
        else:
            sep = '-' if _get_current_date().year - int(century + year) < 100 else '+'
        
        self.century = century
        self.full_year = century + year
        self.year = year
        self.month = month
        self.day = day
        self.sep = sep
        self.num = num
        self.check = check

    def _validate(self):
        """
        Validate a Swedish personal identity number
        """
        if len(self.check) == 0:
            raise PersonnummerInvalidException

        is_valid = _luhn(self.year + self.month + self.day + self.num) == int(self.check)
        if not is_valid:
            raise PersonnummerInvalidException

        try:
            self.get_date()
        except ValueError:
            raise PersonnummerInvalidException

    @staticmethod
    def parse(ssn, options=None):
        """
        Returns a new Personnummer object
        :param ssn
        :type ssn str/int
        :param options
        :type options dict
        :rtype: Personnummer
        :return:
        """
        return Personnummer(ssn, options)


def _luhn(data):
    """
    Calculates the Luhn checksum of a string of digits
    :param data
    :type data str
    :rtype: int
    :return:
    """
    calculation = 0

    for i in range(len(data)):
        v = int(data[i])
        v *= 2 - (i % 2)
        if v > 9:
            v -= 9
        calculation += v

    return int(math.ceil(float(calculation) / 10) * 10 - float(calculation))


def parse(ssn, options=None):
    """
    Returns a new Personnummer object
    :param ssn
    :type ssn str/int
    :param options
    :type options dict
    :rtype: Personnummer
    :return:
    """
    return Personnummer.parse(ssn, options)


def valid(ssn):
    """
    Checks if a ssn is a valid Swedish personal identity number
    :param ssn A Swedish personal identity number
    :type ssn str/int
    """
    try:
        parse(ssn)
        return True
    except PersonnummerException:
        return False


def _get_current_date():
    """
    Get current time. The purpose of this function is to be able to mock
    current time during tests

    :return:
    :rtype datetime.datetime:
    """
    return datetime.date.today()
